"""
    symbolite.abstract
    ~~~~~~~~~~~~~~~~~~

    Abstract symbolite primitives.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .scalar import Scalar
from .symbol import Function, Symbol
from .vector import Vector

__all__ = ["Symbol", "Function", "Scalar", "Vector"]



from types import ModuleType
from typing import Any

import pytest

from symbolite import scalar
from symbolite.core.util import eval_content, substitute_content
from symbolite.impl import libstd


class SimpleVariable(scalar.Scalar):
    """Special type of Scalar that is evaluated to itself."""

    def __repr__(self):
        return self.name

    def eval(self, libsl: ModuleType | None = None):
        return self


class SimpleParameter(scalar.Scalar):
    """Special type of Scalar that is evaluated to itself."""

    def __repr__(self):
        return self.name

    def eval(self, libsl: ModuleType | None = None):
        return self


def is_dependency(x: Any) -> bool:
    return isinstance(x, (SimpleParameter, SimpleVariable))


def test_substitute_content():
    d = {SimpleParameter("x"): 1, SimpleParameter("y"): 2 * SimpleParameter("x")}

    assert substitute_content(d, is_dependency=is_dependency) == {
        SimpleParameter("x"): 1,
        SimpleParameter("y"): (2 * SimpleParameter("x")).subs(
            {SimpleParameter("x"): 1}
        ),
    }


def test_eval_content():
    d = {SimpleParameter("x"): 1}

    assert eval_content(d, libsl=libstd, is_dependency=is_dependency) == {
        SimpleParameter("x"): 1
    }

    d = {SimpleParameter("x"): 1, SimpleParameter("y"): 2 * SimpleParameter("x")}

    assert eval_content(d, libsl=libstd, is_dependency=is_dependency) == {
        SimpleParameter("x"): 1,
        SimpleParameter("y"): 2,
    }


def test_cyclic():
    d = {
        SimpleParameter("x"): 2 * SimpleParameter("x"),
    }

    with pytest.raises(ValueError):
        assert eval_content(d, libsl=libstd, is_dependency=is_dependency)

    d = {
        SimpleParameter("x"): 2 * SimpleParameter("y"),
        SimpleParameter("y"): 2 * SimpleParameter("x"),
    }

    with pytest.raises(ValueError):
        assert eval_content(d, libsl=libstd, is_dependency=is_dependency)



"""
    symbolite.impl.libsympy
    ~~~~~~~~~~~~~~~~~~~~~~~

    Translate symbolite
    into values and functions defined in SymPy.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from . import scalar, symbol, vector

__all__ = ["symbol", "scalar", "vector"]



"""
    symbolite.impl.libsympy.scalar
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Translate symbolite.abstract.scalar
    into values and functions defined in SymPy.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import sympy as sy
from sympy.abc import x, y

from symbolite.core import Unsupported

abs = sy.Abs
acos = sy.acos
acosh = sy.acosh
asin = sy.asin
asinh = sy.asinh
atan = sy.atan
atan2 = sy.atan2
atanh = sy.atanh
ceil = sy.ceiling
comb = Unsupported
copysign = Unsupported
cos = sy.cos
cosh = sy.cosh
degrees = sy.Lambda(x, x * 180 / sy.pi)
erf = Unsupported
erfc = Unsupported
exp = sy.exp
expm1 = sy.Lambda(x, sy.exp(x) - 1)
fabs = sy.Abs
factorial = Unsupported
floor = sy.floor
fmod = sy.Mod
frexp = Unsupported
gamma = Unsupported
hypot = sy.Lambda((x, y), sy.sqrt(x * x + y * y))
isfinite = Unsupported
isinf = Unsupported
isnan = Unsupported
isqrt = Unsupported
ldexp = Unsupported
lgamma = sy.loggamma
log = sy.log
log10 = sy.Lambda(x, sy.log(x, 10))
log1p = sy.Lambda(x, sy.log(1 + x))
log2 = sy.Lambda(x, sy.log(x, 2))
modf = Unsupported
nextafter = Unsupported
pow = sy.Pow
radians = sy.Lambda(x, x * sy.pi / 180)
remainder = Unsupported
sin = sy.sin
sinh = sy.sinh
sqrt = sy.sqrt
tan = sy.tan
tanh = sy.tanh
trunc = Unsupported
ulp = Unsupported

e = sy.exp(1)
inf = sy.oo
pi = sy.pi
nan = sy.nan
tau = 2 * sy.pi

Scalar = sy.Symbol

del sy, Unsupported, x, y



"""
    symbolite.abstract.scalar
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Objects and functions for scalar operations.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import dataclasses
from typing import Any

from ..core import Unsupported
from .symbol import BaseFunction, Symbol

NumberT = int | float | complex


@dataclasses.dataclass(frozen=True)
class Scalar(Symbol):
    """A user defined symbol."""

    def __getitem__(self, key: Any):
        return Unsupported

    def __getattr__(self, key: Any):
        raise AttributeError(key)


@dataclasses.dataclass(frozen=True)
class ScalarUnaryFunction(BaseFunction):
    namespace: str = "scalar"
    arity: int = 1

    def __call__(self, arg1: Scalar | NumberT) -> Scalar:
        return super()._call(arg1)  # type: ignore


@dataclasses.dataclass(frozen=True)
class ScalarBinaryFunction(BaseFunction):
    namespace: str = "scalar"
    arity: int = 2

    def __call__(self, arg1: Scalar | NumberT, arg2: Scalar | NumberT) -> Scalar:
        return super()._call(arg1, arg2)  # type: ignore


# "gcd": None,  # 1 to ---
# "hypot": None,  # 1 to ---
# "isclose": None,  # 2, 3, 4
# "lcm": None,  # 1 to ---
# "perm": None,  # 1 or 2
# "log": None,  # 1 or 2 is used as log(x, e)

abs = ScalarUnaryFunction("abs")
acos = ScalarUnaryFunction("acos")
acosh = ScalarUnaryFunction("acosh")
asin = ScalarUnaryFunction("asin")
asinh = ScalarUnaryFunction("asinh")
atan = ScalarUnaryFunction("atan")
atan2 = ScalarBinaryFunction("atan2")
atanh = ScalarUnaryFunction("atanh")
ceil = ScalarUnaryFunction("ceil")
comb = ScalarUnaryFunction("comb")
copysign = ScalarUnaryFunction("copysign")
cos = ScalarUnaryFunction("cos")
cosh = ScalarUnaryFunction("cosh")
degrees = ScalarUnaryFunction("degrees")
erf = ScalarUnaryFunction("erf")
erfc = ScalarUnaryFunction("erfc")
exp = ScalarUnaryFunction("exp")
expm1 = ScalarUnaryFunction("expm1")
fabs = ScalarUnaryFunction("fabs")
factorial = ScalarUnaryFunction("factorial")
floor = ScalarUnaryFunction("floor")
fmod = ScalarUnaryFunction("fmod")
frexp = ScalarUnaryFunction("frexp")
gamma = ScalarUnaryFunction("gamma")
hypot = ScalarUnaryFunction("gamma")
isfinite = ScalarUnaryFunction("isfinite")
isinf = ScalarUnaryFunction("isinf")
isnan = ScalarUnaryFunction("isnan")
isqrt = ScalarUnaryFunction("isqrt")
ldexp = ScalarBinaryFunction("ldexp")
lgamma = ScalarUnaryFunction("lgamma")
log = ScalarUnaryFunction("log")
log10 = ScalarUnaryFunction("log10")
log1p = ScalarUnaryFunction("log1p")
log2 = ScalarUnaryFunction("log2")
modf = ScalarUnaryFunction("modf")
nextafter = ScalarUnaryFunction("nextafter")
pow = ScalarUnaryFunction("pow")
radians = ScalarUnaryFunction("radians")
remainder = ScalarBinaryFunction("remainder")
sin = ScalarUnaryFunction("sin")
sinh = ScalarUnaryFunction("sinh")
sqrt = ScalarUnaryFunction("sqrt")
tan = ScalarUnaryFunction("tan")
tanh = ScalarUnaryFunction("tanh")
tan = ScalarUnaryFunction("tan")
trunc = ScalarUnaryFunction("trunc")
ulp = ScalarUnaryFunction("ulp")

e = Scalar("e", namespace="scalar")
inf = Scalar("inf", namespace="scalar")
pi = Scalar("pi", namespace="scalar")
nan = Scalar("nan", namespace="scalar")
tau = Scalar("tau", namespace="scalar")

del BaseFunction, Symbol, dataclasses



from symbolite import Symbol
from symbolite.core.symbolgroup import AutoSymbol, SymbolicNamespace


def test_naming():
    class N(SymbolicNamespace):
        x = AutoSymbol()
        y = AutoSymbol()

    assert isinstance(N.x, Symbol)
    assert isinstance(N.y, Symbol)
    assert N.x.name == "x"
    assert N.y.name == "y"
    assert N.symbol_names() == {"x", "y"}


def test_eq():
    class N(SymbolicNamespace):
        x = AutoSymbol()
        y = AutoSymbol()

        x.eq(2 * y)



from __future__ import annotations

import dataclasses
import types
from typing import Any, Iterable, Mapping

from ..abstract.symbol import Symbol


class SymbolicList(list[Symbol]):
    @classmethod
    def from_iterable(cls, it: Iterable[Symbol]):
        return cls(it)

    def subs(self, *mappers: Mapping[Any, Any]) -> SymbolicList:
        """Replace symbols, functions, values, etc by others.

        If multiple mappers are provided,
            they will be used in order (using a ChainMap)

        If a given object is not found in the mappers,
            the same object will be returned.

        Parameters
        ----------
        *mappers
            dictionaries mapping source to destination objects.
        """
        return self.__class__.from_iterable((se.subs(*mappers) for se in self))

    def subs_by_name(self, **symbols: Any) -> SymbolicList:
        """Replace Symbols by values or objects, matching by name.

        If multiple mappers are provided,
            they will be used in order (using a ChainMap)

        If a given object is not found in the mappers,
            the same object will be returned.

        Parameters
        ----------
        **symbols
            keyword arguments connecting names to values.
        """
        return self.__class__.from_iterable((se.subs_by_name(**symbols) for se in self))

    def eval(self, **libs: types.ModuleType) -> SymbolicList:
        """Evaluate expression.

        If no implementation library is provided:
        1. 'libsl' will be looked up going back though the stack
           until is found.
        2. If still not found, the implementation using the python
           math module will be used (and a warning will be issued).

        Parameters
        ----------
        libs
            implementations
        """

        return self.__class__.from_iterable(se.eval(**libs) for se in self)

    def symbol_names(self, namespace: str | None = "") -> set[str]:
        """Return a set of symbol names (with full namespace indication).

        Parameters
        ----------
        namespace: str or None
            If None, all symbols will be returned independently of the namespace.
            If a string, will compare Symbol.namespace to that.
            Defaults to "" which is the namespace for user defined symbols.
        """
        s: set[str] = set()
        return s.union(*(se.symbol_names(namespace) for se in self))

    def __str__(self):
        return "\n".join(str(se) for se in self)


class SymbolicNamespace:
    expressions: SymbolicList = SymbolicList()

    @classmethod
    def symbol_names(cls, namespace: str | None = "") -> set[str]:
        return cls.expressions.symbol_names(namespace)


@dataclasses.dataclass(frozen=True)
class AutoSymbol(Symbol):
    name: str = "<auto>"

    def __set_name__(self, owner: Any, name: str):
        if issubclass(owner, SymbolicNamespace):
            object.__setattr__(self, "name", name)
            owner.expressions.append(self)



import importlib
import inspect
import types
from pathlib import Path


def find_module_in_stack(name: str = "libsl") -> types.ModuleType | None:
    """Find libraries in stack.

    Parameters
    ----------
    expr
        If None, an implementation for every abstract library
        will be look for.
        If an expression, it will be first inspected to find
        which libraries it is using and only those will be look for.

    """
    frame = inspect.currentframe()
    while frame:
        if name in frame.f_locals:
            mod = frame.f_locals[name]
            if mod is not None:
                return mod
        frame = frame.f_back

    return None


def get_all_implementations() -> dict[str, types.ModuleType]:
    out = {}

    path = Path(__file__)
    for p in path.parent.iterdir():
        name = p.stem
        if name.startswith("_"):
            continue

        try:
            module = importlib.import_module(f".{name}", package=__package__)
        except ImportError:
            pass
        else:
            out[name] = module

    return out



"""
    symbolite
    ~~~~~~~~~

    A minimal symbolic python package.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from .abstract import Function, Scalar, Symbol, Vector, scalar, vector

__all__ = ["Symbol", "Function", "Scalar", "scalar", "Vector", "vector"]



"""
    symbolite.core
    ~~~~~~~~~~~~~~

    Symbolite core classes and functions.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import collections
import types
import warnings
from typing import Any, Callable, Mapping, Sequence

from ..impl import find_module_in_stack


class Unsupported(ValueError):
    """Label unsupported"""


def as_string(expr: Any) -> str:
    """Return the expression as a string.

    Parameters
    ----------
    expr
        symbolic expression.
    """
    return str(expr)


def as_function(
    expr: Any,
    function_name: str,
    params: Sequence[str],
    libsl: types.ModuleType | None = None,
) -> Callable[..., Any]:
    """Converts the expression to a callable function.

    Parameters
    ----------
    expr
        symbolic expression.
    function_name
        name of the function to be used.
    params
        names of the parameters.
    libsl
        implementation module.
    """

    function_def = (
        f"""def {function_name}({", ".join(params)}): return {as_string(expr)}"""
    )

    lm = compile(function_def, libsl)

    return lm[function_name]


def compile(
    code: str,
    libsl: types.ModuleType | None = None,
) -> dict[str, Any]:
    """Compile the code and return the local dictionary.

    Parameters
    ----------
    expr
        symbolic expression.
    libsl
        implementation module.
    """

    if libsl is None:
        libsl = find_module_in_stack()
    if libsl is None:
        warnings.warn("No libsl provided, defaulting to Python standard library.")
        from ..impl import libstd as libsl

    assert libsl is not None

    lm: dict[str, Any] = {}
    exec(
        code,
        {
            "symbol": libsl.symbol,
            "scalar": libsl.scalar,
            "vector": libsl.vector,
            **globals(),
        },
        lm,
    )
    return lm


def inspect(expr: Any) -> dict[Any, int]:
    """Inspect an expression and return what is there
    and how many times.

    Parameters
    ----------
    expr
        symbolic expression.
    """
    if hasattr(expr, "yield_named"):
        cnt = collections.Counter[Any](expr.yield_named())
        return dict(cnt)
    return {expr: 1}


def evaluate(expr: Any, libsl: types.ModuleType | None = None) -> Any:
    """Evaluate expression.

    Parameters
    ----------
    expr
        symbolic expression.
    libsl
        implementation module.
    """

    if libsl is None:
        libsl = find_module_in_stack()
    if libsl is None:
        warnings.warn("No libsl provided, defaulting to Python standard library.")
        from ..impl import libstd as libsl

    if hasattr(expr, "eval"):
        return expr.eval(libsl)
    return expr


def substitute(expr: Any, replacements: Mapping[Any, Any]) -> Any:
    """Replace symbols, functions, values, etc by others.

    Parameters
    ----------
    expr
        symbolic expression.
    replacements
        replacement dictionary.
    """
    if hasattr(expr, "subs"):
        return expr.subs(replacements)
    return replacements.get(expr, expr)


def substitute_by_name(expr: Any, **replacements: Any) -> Any:
    """Replace Symbols by values or objects, matching by name.

    Parameters
    ----------
    expr
        symbolic expression.
    replacements
        replacement dictionary.
    """

    if hasattr(expr, "subs_by_name"):
        return expr.subs_by_name(**replacements)
    return expr



"""
    symbolite.impl.libstd.scalar
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Translate symbolite.abstract.scalar
    into values and functions in Python standard library.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import math

from ...core import Unsupported

# "gcd": None,  # 1 to ---
# "hypot": None,  # 1 to ---
# "isclose": None,  # 2, 3, 4
# "lcm": None,  # 1 to ---
# "perm": None,  # 1 or 2
# "log": None,  # 1 or 2 is used as log(x, e)

abs = abs
acos = math.acos
acosh = math.acosh
asin = math.asin
asinh = math.asinh
atan = math.atan
atan2 = math.atan2
atanh = math.atanh
ceil = math.ceil
comb = math.comb
copysign = math.copysign
cos = math.cos
cosh = math.cosh
degrees = math.degrees
erf = math.erf
erfc = math.erfc
exp = math.exp
expm1 = math.expm1
fabs = math.fabs
factorial = math.factorial
floor = math.floor
fmod = math.fmod
frexp = math.frexp
gamma = math.gamma
hypot = math.hypot
isfinite = math.isfinite
isinf = math.isinf
isnan = math.isnan
isqrt = math.isqrt
ldexp = math.ldexp
lgamma = math.lgamma
log = math.log
log10 = math.log10
log1p = math.log1p
log2 = math.log2
modf = math.modf
nextafter = math.nextafter
log2 = math.log2
pow = math.pow
radians = math.radians
remainder = math.remainder
sin = math.sin
sinh = math.sinh
sqrt = math.sqrt
tan = math.tan
tanh = math.tanh
tan = math.tan
trunc = math.trunc
ulp = math.ulp

e = math.e
inf = math.inf
pi = math.pi
nan = math.nan
tau = math.tau

Scalar = Unsupported

del math, Unsupported



"""
    symbolite.impl.libnumpy.vector
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Translate symbolite.abstract.vector
    into values and functions defined in NumPy.

    :copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import operator

import numpy as np

from symbolite.core import Unsupported

op_getitem = operator.getitem

sum = np.sum
prod = np.prod

Vector = Unsupported


del np, operator, Unsupported
